@@ -10,10 +10,14 @@ setAs("Matrix", "denseMatrix",  function(from) as_dense(from))
 ## Maybe TODO:
 ## setAs("Matrix", "nMatrix", function(from) ....)
 
+## Anything: we build on  as.matrix(.) :
+## ---       authors can always provide their own specific  setAs(*, "Matrix")
+setAs("ANY", "Matrix", function(from) Matrix(as.matrix(from)))
+
 ## Most of these work; this is a last resort:
-setAs(from = "Matrix", to = "matrix", # do *not* call base::as.matrix() here:
+setAs("Matrix", "matrix", # do *not* call base::as.matrix() here:
       function(from) .bail.out.2("coerce", class(from), class(to)))
-setAs(from = "matrix", to = "Matrix", function(from) Matrix(from))
+setAs("matrix", "Matrix", function(from) Matrix(from))
 
 ## ## probably not needed eventually:
 ## setAs(from = "ddenseMatrix", to = "matrix",
@@ -37,9 +41,10 @@ setMethod("drop", signature(x = "Matrix"),
 	  function(x) if(all(dim(x) != 1)) x else drop(as(x, "matrix")))
 
 ## slow "fall back" method {subclasses should have faster ones}:
-setMethod("as.vector", signature(x = "Matrix", mode = "missing"),
+setMethod("as.vector", "Matrix",
 	  function(x, mode) as.vector(as(x, "matrix"), mode))
 ## so base functions calling as.vector() work too:
+## S3 dispatch works for base::as.vector(), but S4 dispatch does not
 as.vector.Matrix <- function(x, mode) as.vector(as(x, "matrix"), mode)
 
 setAs("Matrix", "vector",  function(from) as.vector (as(from, "matrix")))
@@ -110,18 +115,17 @@ setMethod("isTriangular", signature(object = "matrix"), isTriMat)
 setMethod("isDiagonal", signature(object = "matrix"), .is.diagonal)
 
 ## The "catch all" methods -- far from optimal:
-setMethod("symmpart", signature(x = "Matrix"),
-	  function(x) as((x + t(x))/2, "symmetricMatrix"))
-setMethod("skewpart", signature(x = "Matrix"),
-	  function(x) (x - t(x))/2)
+setMethod("symmpart", signature(x = "Matrix"), function(x)
+    as(symmetrizeDimnames(x + t(x))/2, "symmetricMatrix"))
+setMethod("skewpart", signature(x = "Matrix"), function(x) symmetrizeDimnames(x - t(x))/2)
 
 ## FIXME: do this (similarly as for "ddense.." in C
-setMethod("symmpart", signature(x = "matrix"), function(x) (x + t(x))/2)
-setMethod("skewpart", signature(x = "matrix"), function(x) (x - t(x))/2)
+setMethod("symmpart", signature(x = "matrix"), function(x) symmetrizeDimnames(x + t(x))/2)
+setMethod("skewpart", signature(x = "matrix"), function(x) symmetrizeDimnames(x - t(x))/2)
 
 
-if(getRversion() > "3.1.0")
-## NB: "nsparseMatrix" has extra method
+if(getRversion() >= "3.1.0")
+## NB: ./nsparseMatrix.R and ./sparseVector.R have extra methods
 setMethod("anyNA", signature(x = "xMatrix"),
 	  function(x) anyNA(x@x))
 
@@ -143,19 +147,30 @@ dimnamesGets <- function (x, value) {
 	!(is.null(v2 <- value[[2]]) || length(v2) == d[2]))
 	stop(gettextf("invalid dimnames given for %s object", dQuote(class(x))),
 	     domain=NA)
-    x@Dimnames <- list(if(!is.null(v1)) as.character(v1),
-		       if(!is.null(v2)) as.character(v2))
+    x@Dimnames <- .fixupDimnames(value)
     x
 }
-setMethod("dimnames<-", signature(x = "Matrix", value = "list"),
-	  dimnamesGets)
+dimnamesGetsNULL <- function(x) {
+    message("dimnames(.) <- NULL:  translated to \ndimnames(.) <- list(NULL,NULL)  <==>  unname(.)")
+    x@Dimnames <- list(NULL,NULL)
+    x
+}
+setMethod("dimnames<-", signature(x = "compMatrix", value = "list"),
+          function(x, value) { ## "compMatrix" have 'factors' slot
+              if(length(x@factors)) x@factors <- list()
+              dimnamesGets(x, value)
+          })
+setMethod("dimnames<-", signature(x = "Matrix", value = "list"), dimnamesGets)
+
+setMethod("dimnames<-", signature(x = "compMatrix", value = "NULL"),
+          function(x, value) { ## "compMatrix" have 'factors' slot
+              if(length(x@factors)) x@factors <- list()
+              dimnamesGetsNULL(x)
+          })
 
 setMethod("dimnames<-", signature(x = "Matrix", value = "NULL"),
-	  function(x, value) {
-	      message("dimnames(.) <- NULL:  translated to \ndimnames(.) <- list(NULL,NULL)  <==>  unname(.)")
-	      x@Dimnames <- list(NULL,NULL)
-	      x
-	  })
+	  function(x, value) dimnamesGetsNULL(x))
+
 
 setMethod("unname", signature("Matrix", force="missing"),
 	  function(obj) { obj@Dimnames <- list(NULL,NULL); obj})
@@ -165,8 +180,6 @@ Matrix <- function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
                     dimnames = NULL, sparse = NULL,
                     doDiag = TRUE, forceCheck = FALSE)
 {
-    sparseDefault <- function(m) prod(dim(m)) > 2*sum(isN0(as(m, "matrix")))
-
     i.M <- is(data, "Matrix")
     sM <- FALSE
     if(i.M) {
@@ -181,6 +194,7 @@ Matrix <- function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
     if(is.null(sparse1 <- sparse) && (i.M || is(data, "matrix")))
 	sparse <- sparseDefault(data)
     doDN <- TRUE
+    i.m <- is.matrix(data)
     if (i.M) {
 	if (!sV) {
 	    if(!missing(nrow) || !missing(ncol)|| !missing(byrow))
@@ -191,7 +205,7 @@ Matrix <- function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
 	    ## else : convert  dense <-> sparse -> at end
 	}
     }
-    else if (!is.matrix(data)) { ## cut & paste from "base::matrix" :
+    else if (!i.m) { ## cut & paste from "base::matrix" :
 	## avoid copying to strip attributes in simple cases
 	if (is.object(data) || !is.atomic(data)) data <- as.vector(data)
 	if(length(data) == 1 && is0(data) && !identical(sparse, FALSE)) {
@@ -220,8 +234,8 @@ Matrix <- function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
 	    if(is.null(sparse))
 		sparse <- sparseDefault(data)
 	}
-        doDN <- FALSE
-    } else if(!missing(nrow) || !missing(ncol)|| !missing(byrow))
+        doDN <- FALSE # .. set above
+    } else if(!missing(nrow) || !missing(ncol)|| !missing(byrow)) ## i.m == is.matrix(.)
 	warning("'nrow', 'ncol', etc, are disregarded for matrix 'data'")
 
     ## 'data' is now a "matrix" or "Matrix"
@@ -294,12 +308,12 @@ setMethod("chol2inv", signature(x = "denseMatrix"),
 	  function (x, ...) chol2inv(as(as(x, "dMatrix"), "dtrMatrix"), ...))
 setMethod("chol2inv", signature(x = "diagonalMatrix"),
 	  function (x, ...) {
-	      chk.s(...)
+	      chk.s(..., which.call=-2)
 	      tcrossprod(solve(x))
 	  })
 setMethod("chol2inv", signature(x = "sparseMatrix"),
 	  function (x, ...) {
-	      chk.s(...)
+	      chk.s(..., which.call=-2)
 	      ## for now:
 	      tcrossprod(solve(as(x,"triangularMatrix")))
 	  })
@@ -422,92 +436,6 @@ setMethod("image", "Matrix",
 
 ## Group Methods
 
-## --- "Summary" ------- have "ddense*" and "dsparse*" ones in ---> ./dMatrix.R <---
-##      -------          "diagMatrix" --> ./diagMatrix.R            ~~~~~~~~~~~
-## For all other Matrix objects {and note that  "all" and "any" have their own}:
-
-setMethod("all", signature(x = "Matrix"),
-	  function(x, ..., na.rm)
-	  callGeneric(as(x, "lMatrix"), ..., na.rm=na.rm))
-
-setMethod("any", signature(x = "Matrix"),
-	  function(x, ..., na.rm)
-	  callGeneric(as(x, "lMatrix"), ..., na.rm=na.rm))
-
-setMethod("Summary", signature(x = "Matrix", na.rm = "ANY"),
-	  function(x, ..., na.rm)
-	  callGeneric(as(x,"dMatrix"), ..., na.rm = na.rm))
-
-## Try to make   min(1, <Matrix>)  work, i.e., not dispatch on first arg to .Primitive
-## This for(..) gives {during installation}
-## Error in setGeneric(F, signature = "...") :
-##   ‘max’ is a primitive function;  methods can be defined, but the generic function is implicit, and cannot be changed.
-if(FALSE)
-for(F in c("max", "min", "range", "prod", "sum", "any", "all")) {
-    setGeneric(F, signature = "...")
-}
-### try on "min" for now --- ~/R/Pkgs/Rmpfr/R/mpfr.R is the example (for "pmin")
-if(FALSE)### This gives error message that the "ANY" is method is sealed ...
-setMethod("min", "ANY",
-	  function(..., na.rm = FALSE) {
-	      args <- list(...)
-	      if(all(isAtm <- vapply(args, is.atomic, NA)))
-		  return( base::min(..., na.rm = na.rm) )
-              ## else try to dispatch on an argument which is a Matrix.. or in a
-              if(any(isM <- vapply(args, is, NA, class2="Matrix"))) {
-                  ## swap the Matrix with the first argument
-                  i <- which.max(isM)# the first "Matrix"
-                  if(i == 1)
-                      stop("programming error: min() should have dispatched w/ 1st arg much earlier")
-              } else { ## if no "Matrix", take the first non-atomic argument
-                  ## (FIXME: should take the first for which there is a method !)
-                  i <- which.max(!isAtm)
-              }
-              ii <- seq_along(args)
-              ii[c(1,i)] <- c(i,1)
-              do.call(min, c(args[ii], list(na.rm=na.rm)))
-          })
-
-if(FALSE) { ## FIXME: it does *not* solve the problem anyway ..
-##
-##  (m <- Matrix(c(0,0,2:0), 3,5))
-##   min(1,m)
-##-> error, as it calls the .Primitive min() and that does *not* dispatch on 2nd arg
-##
-setMethod("Summary", signature(x = "ANY", na.rm = "ANY"),
-	  function(x, ..., na.rm) {
-          if(!length(a <- list(...))) (get(.Generic, envir=baseenv()))(x, na.rm=na.rm)
-          else {
-                 if(!is.null(v <- getOption("Matrix.verbose")) && v >= 1)
-		     message(sprintf("in Summary(<ANY>, .): %s(<%s>, <%s>%s)\n",
-				     .Generic, class(x), class(a[[1]]), if(length(a) > 1)", ..." else ""),
-			     domain = NA)
-                 do.call(.Generic, c(x, a, list(na.rm=na.rm)))
-             }})
-}
-
-Summary.l <- function(x, ..., na.rm) { ## must be method directly
-    if(.Generic %in% c("all", "any"))
-	callGeneric(x@x, ..., na.rm = na.rm)
-    else {
-	r <- callGeneric(as(x,"dMatrix"), ..., na.rm = na.rm)
-	if(!is.infinite(r) && .Generic != "prod") as.integer(r) else r
-    }
-}
-## almost identical:
-Summary.np <- function(x, ..., na.rm) {
-    if(.Generic %in% c("all", "any"))
-	callGeneric(as(x, "lMatrix"), ..., na.rm = na.rm)
-    else {
-	r <- callGeneric(as(x,"dMatrix"), ..., na.rm = na.rm)
-	if(!is.infinite(r) && .Generic != "prod") as.integer(r) else r
-    }
-}
-##
-setMethod("Summary", signature(x = "lMatrix", na.rm = "ANY"), Summary.l)
-setMethod("Summary", signature(x = "nMatrix", na.rm = "ANY"), Summary.np)
-setMethod("Summary", signature(x = "indMatrix", na.rm = "ANY"), Summary.np)
-
 ## NOTE:  "&" and "|"  are now in group "Logic" c "Ops" --> ./Ops.R
 ##        "!" is in ./not.R
 
@@ -581,7 +509,7 @@ setMethod("[", signature(x = "Matrix", i = "ANY", j = "ANY", drop = "ANY"),
 
 	## Note: current method dispatch seems not to call this ever
 
-	if(!any(is.na(i)) && all(i)) ## select everything
+	if(!anyNA(i) && all(i)) ## select everything
 	    x
 	else ## not selecting all -> result is *NOT* diagonal/triangular/symmetric/..
 	    ## keep j missing, but  drop = "logical"
@@ -629,8 +557,8 @@ subset.ij <- function(x, ij) {
 		ij.x <- non0.i(x, cld)
 	    }
 
-	    m1 <- .Call(m_encodeInd, ij.x, di, checkBounds = FALSE)
-            m2 <- .Call(m_encodeInd, ij -1L, di, checkBounds = TRUE)
+	    m1 <- .Call(m_encodeInd, ij.x, di, orig1=FALSE, checkBounds=FALSE)
+            m2 <- .Call(m_encodeInd, ij,   di, orig1= TRUE, checkBounds= TRUE)
 	    mi <- match(m2, m1, nomatch=0)
 	    mmi <- mi != 0L ## == (m2 %in% m1)
 	    ## Result: all FALSE or 0  apart from where we match non-zero entries
@@ -706,7 +634,7 @@ setMethod("[", signature(x = "Matrix", i = "matrix", j = "missing", drop="missin
 	if(!is.integer(i)) storage.mode(i) <- "integer"
 	if(any(i < 0))
 	    stop("negative values are not allowed in a matrix subscript")
-	if(any(is.na(i)))
+	if(anyNA(i))
 	    stop("NAs are not allowed in subscripted assignments")
 	if(any(i0 <- (i == 0))) # remove them
             i <- i[ - which(i0, arr.ind = TRUE)[,"row"], ]
@@ -745,7 +673,10 @@ setReplaceMethod("[", signature(x = "Matrix", i = "missing", j = "ANY",
 setReplaceMethod("[", signature(x = "Matrix", i = "ANY", j = "missing",
 				value = "Matrix"),
 		 function (x, i, j, ..., value)
-		 callGeneric(x=x, i=i, , value = as.vector(value)))
+		     if(nargs() == 3)
+			 callGeneric(x=x, i=i, value = as.vector(value))
+		     else
+			 callGeneric(x=x, i=i, , value = as.vector(value)))
 
 setReplaceMethod("[", signature(x = "Matrix", i = "ANY", j = "ANY",
 				value = "Matrix"),
@@ -761,7 +692,10 @@ setReplaceMethod("[", signature(x = "Matrix", i = "missing", j = "ANY",
 setReplaceMethod("[", signature(x = "Matrix", i = "ANY", j = "missing",
 				value = "matrix"),
 		 function (x, i, j, ..., value)
-		 callGeneric(x=x, i=i, , value = c(value)))
+		     if(nargs() == 3)
+			 callGeneric(x=x, i=i, value = c(value))
+		     else
+			 callGeneric(x=x, i=i, , value = c(value)))
 
 setReplaceMethod("[", signature(x = "Matrix", i = "ANY", j = "ANY",
 				value = "matrix"),
