@@ -120,7 +120,7 @@ if(FALSE)##--- no longer used:
     ## block-diagonal matrix [a dgTMatrix] from list of matrices
     stopifnot(is.list(lst), (nl <- length(lst)) >= 1)
 
-    Tlst <- lapply(lapply(lst, Matrix:::as_Csp2), # includes "diagU2N"
+    Tlst <- lapply(lapply(lst, as_Csp2), # includes "diagU2N"
 		   as, "TsparseMatrix")
     if(nl == 1) return(Tlst[[1]])
     ## else
@@ -198,6 +198,13 @@ bdiag <- function(...) {
     as(.bdiag(alis), "CsparseMatrix")
 }
 
+setMethod("tril", "diagonalMatrix", function(x, k = 0, ...)
+    if(k >= 0) x else .setZero(x, paste0(.M.kind(x), "tCMatrix")))
+
+setMethod("triu", "diagonalMatrix", function(x, k = 0, ...)
+    if(k <= 0) x else  .setZero(x, paste0(.M.kind(x), "tCMatrix")))
+
+
 
 .diag2tT <- function(from, uplo = "U", kind = .M.kind(from)) {
     ## to triangular Tsparse
@@ -227,7 +234,7 @@ bdiag <- function(...) {
 		n))
 }
 
-## diagonal -> triangular,  upper / lower depending on "partner":
+## diagonal -> triangular,  upper / lower depending on "partner" 'x':
 diag2tT.u <- function(d, x, kind = .M.kind(d))
     .diag2tT(d, uplo = if(is(x,"triangularMatrix")) x@uplo else "U", kind)
 
@@ -251,33 +258,30 @@ setMethod("is.infinite", signature(x = "diagonalMatrix"),
 ## "hack"  instead of signature  x = "diagonalMatrix" :
 ##
 ## ddi*:
-diag2tT <- function(from) .diag2tT(from, "U", "d")
-setAs("ddiMatrix", "triangularMatrix", diag2tT)
-##_no_longer_ setAs("ddiMatrix", "sparseMatrix", diag2tT)
+di2tT <- function(from) .diag2tT(from, "U", "d")
+setAs("ddiMatrix", "triangularMatrix", di2tT)
+##_no_longer_ setAs("ddiMatrix", "sparseMatrix", di2tT)
 ## needed too (otherwise <dense> -> Tsparse is taken):
-setAs("ddiMatrix", "TsparseMatrix", diag2tT)
-setAs("ddiMatrix", "dsparseMatrix", diag2tT)
+setAs("ddiMatrix", "TsparseMatrix", di2tT)
+setAs("ddiMatrix", "dsparseMatrix", di2tT)
 setAs("ddiMatrix", "CsparseMatrix",
       function(from) as(.diag2tT(from, "U", "d"), "CsparseMatrix"))
-setAs("ddiMatrix", "symmetricMatrix",
-      function(from) .diag2sT(from, "U", "d"))
+setAs("ddiMatrix", "symmetricMatrix", function(from) .diag2sT(from, "U", "d"))
 ##
 ## ldi*:
-diag2tT <- function(from) .diag2tT(from, "U", "l")
-setAs("ldiMatrix", "triangularMatrix", diag2tT)
-##_no_longer_ setAs("ldiMatrix", "sparseMatrix", diag2tT)
+di2tT <- function(from) .diag2tT(from, "U", "l")
+setAs("ldiMatrix", "triangularMatrix", di2tT)
+##_no_longer_ setAs("ldiMatrix", "sparseMatrix", di2tT)
 ## needed too (otherwise <dense> -> Tsparse is taken):
-setAs("ldiMatrix", "TsparseMatrix", diag2tT)
-setAs("ldiMatrix", "lsparseMatrix", diag2tT)
+setAs("ldiMatrix", "TsparseMatrix", di2tT)
+setAs("ldiMatrix", "lsparseMatrix", di2tT)
 setAs("ldiMatrix", "CsparseMatrix",
       function(from) as(.diag2tT(from, "U", "l"), "CsparseMatrix"))
-setAs("ldiMatrix", "symmetricMatrix",
-      function(from) .diag2sT(from, "U", "l"))
-
+setAs("ldiMatrix", "symmetricMatrix", function(from) .diag2sT(from, "U", "l"))
+rm(di2tT)
 
 setAs("diagonalMatrix", "nMatrix",
-      function(from) {
-	  n <- from@Dim[1]
+      di2nMat <- function(from) {
 	  i <- if(from@diag == "U") integer(0) else which(isN0(from@x)) - 1L
 	  new("ntTMatrix", i = i, j = i, diag = from@diag,
 	      Dim = from@Dim, Dimnames = from@Dimnames)
@@ -285,21 +289,27 @@ setAs("diagonalMatrix", "nMatrix",
 
 setAs("diagonalMatrix", "nsparseMatrix", function(from) as(from, "nMatrix"))
 
-## Cheap fast substitute for diag() which *does* preserve the mode of x :
+##' A version of diag(x,n) which *does* preserve the mode of x, where diag() "fails"
 mkDiag <- function(x, n) {
     y <- matrix(as0(mod=mode(x)), n,n)
     if (n > 0) y[1L + 0:(n - 1L) * (n + 1)] <- x
     y
 }
+## NB: diag(x,n) is really faster for n >= 20, and even more for large n
+## --> using diag() where possible, ==> .ddi2mat()
 
-setAs("diagonalMatrix", "matrix",
-      function(from) {
-	  ## want "ldiMatrix" -> <logical> "matrix" :
-	  mkDiag(if(from@diag == "U") as1(from@x) else from@x,
-		 n = from@Dim[1])
-      })
+.diag2mat <- function(from)
+    ## want "ldiMatrix" -> <logical> "matrix"  (but integer -> <double> for now)
+    mkDiag(if(from@diag == "U") as1(from@x) else from@x, n = from@Dim[1])
 
-setMethod("as.vector", signature(x = "diagonalMatrix", mode="missing"),
+.ddi2mat <- function(from)
+    base::diag(if(from@diag == "U") as1(from@x) else from@x, nrow = from@Dim[1])
+
+setAs("ddiMatrix", "matrix", .ddi2mat)
+## the non-ddi diagonalMatrix -- only "ldiMatrix" currently:
+setAs("diagonalMatrix", "matrix", .diag2mat)
+
+setMethod("as.vector", "diagonalMatrix",
 	  function(x, mode) {
 	      n <- x@Dim[1]
               mod.x <- mode(x@x)
@@ -307,7 +317,7 @@ setMethod("as.vector", signature(x = "diagonalMatrix", mode="missing"),
 	      if(n)
 		  r[1 + 0:(n - 1L) * (n + 1)] <-
 		      if(x@diag == "U") as1(mod=mod.x) else x@x
-	      r
+	      as.vector(r, mode)
 	  })
 
 setAs("diagonalMatrix", "generalMatrix", # prefer sparse:
@@ -316,18 +326,19 @@ setAs("diagonalMatrix", "generalMatrix", # prefer sparse:
 setAs("diagonalMatrix", "denseMatrix",
       function(from) as(as(from, "CsparseMatrix"), "denseMatrix"))
 
-.diag.x <- function(m) if(m@diag == "U") rep.int(as1(m@x), m@Dim[1]) else m@x
+..diag.x <- function(m)                   rep.int(as1(m@x), m@Dim[1])
+.diag.x  <- function(m) if(m@diag == "U") rep.int(as1(m@x), m@Dim[1]) else m@x
 
 .diag.2N <- function(m) {
     if(m@diag == "U") m@diag <- "N"
     m
 }
 
-setAs("ddiMatrix", "dgeMatrix",
-      function(from) .Call(dup_mMatrix_as_dgeMatrix, from))
-setAs("ddiMatrix", "ddenseMatrix",
+setAs("ddiMatrix", "dgeMatrix", ..2dge)
+
+setAs("ddiMatrix", "ddenseMatrix", #-> "dtr"
       function(from) as(as(from, "triangularMatrix"),"denseMatrix"))
-setAs("ldiMatrix", "ldenseMatrix",
+setAs("ldiMatrix", "ldenseMatrix", #-> "ltr"
       function(from) as(as(from, "triangularMatrix"),"denseMatrix"))
 
 
@@ -347,7 +358,7 @@ setAs("matrix", "diagonalMatrix",
 	      storage.mode(x) <- "double"
 	  } ## TODO: complex
 	  new(cl, Dim = c(n,n), diag = if(uni) "U" else "N",
-	      x = if(uni) x[FALSE] else x)
+	      x = if(uni) x[FALSE] else x, Dimnames = .M.DN(from))
       })
 
 ## ``generic'' coercion to  diagonalMatrix : build on  isDiagonal() and diag()
@@ -367,7 +378,7 @@ setAs("Matrix", "diagonalMatrix",
 	      storage.mode(x) <- "double"
 	  } ## TODO: complex
 	  new(cl, Dim = c(n,n), diag = if(uni) "U" else "N",
-	      x = if(uni) x[FALSE] else x)
+	      x = if(uni) x[FALSE] else x, Dimnames = from@Dimnames)
       })
 
 
@@ -435,6 +446,27 @@ setReplaceMethod("[", signature(x = "diagonalMatrix", i = "index",
                          replDiag(x, i=i, , value=value)
                  })
 
+setReplaceMethod("[", signature(x = "diagonalMatrix", i = "missing",
+				j = "index", value = "replValue"),
+		 function(x,i,j, ..., value) replDiag(x, j=j, value=value))
+
+## x[] <- value :
+setReplaceMethod("[", signature(x = "diagonalMatrix", i = "missing",
+				j = "missing", value = "ANY"),
+		 function(x,i,j, ..., value)
+	     {
+	      if(all0(value)) { # be faster
+		  r <- new(paste0(.M.kindC(getClassDef(class(x))),"tTMatrix"))# of all "0"
+		  r@Dim <- x@Dim
+		  r@Dimnames <- x@Dimnames
+		  r
+	      } else { ## typically non-sense: assigning to full sparseMatrix
+		  x[TRUE] <- value
+		  x
+	      }
+	  })
+
+
 setReplaceMethod("[", signature(x = "diagonalMatrix",
                                 i = "matrix", # 2-col.matrix
 				j = "missing", value = "replValue"),
@@ -463,10 +495,8 @@ setReplaceMethod("[", signature(x = "diagonalMatrix",
 		     }
 		 })
 
-setReplaceMethod("[", signature(x = "diagonalMatrix", i = "missing",
-				j = "index", value = "replValue"),
-		 function(x,i,j, ..., value) replDiag(x, j=j, value=value))
 
+## value = "sparseMatrix":
 setReplaceMethod("[", signature(x = "diagonalMatrix", i = "missing", j = "index",
 				value = "sparseMatrix"),
 		 function (x, i, j, ..., value)
@@ -480,6 +510,7 @@ setReplaceMethod("[", signature(x = "diagonalMatrix", i = "index", j = "index",
 		 function (x, i, j, ..., value)
 		 callGeneric(x=x, i=i, j=j, value = as(value, "sparseVector")))
 
+## value = "sparseVector":
 setReplaceMethod("[", signature(x = "diagonalMatrix", i = "missing", j = "index",
 				value = "sparseVector"),
 		 replDiag)
@@ -494,15 +525,12 @@ setReplaceMethod("[", signature(x = "diagonalMatrix", i = "index", j = "index",
 setMethod("t", signature(x = "diagonalMatrix"),
           function(x) { x@Dimnames <- x@Dimnames[2:1] ; x })
 
-setMethod("isDiagonal", signature(object = "diagonalMatrix"),
-          function(object) TRUE)
-setMethod("isTriangular", signature(object = "diagonalMatrix"),
-          function(object) TRUE)
-setMethod("isSymmetric", signature(object = "diagonalMatrix"),
-	  function(object, ...) TRUE)
+setMethod("isDiagonal",   "diagonalMatrix", function(object) TRUE)
+setMethod("isTriangular", "diagonalMatrix", function(object, upper=NA, ...) TRUE)
+setMethod("isSymmetric",  "diagonalMatrix", function(object, ...) TRUE)
 
 setMethod("symmpart", signature(x = "diagonalMatrix"), function(x) x)
-setMethod("skewpart", signature(x = "diagonalMatrix"), setZero)
+setMethod("skewpart", signature(x = "diagonalMatrix"), function(x) .setZero(x))
 
 setMethod("chol", signature(x = "ddiMatrix"),
 	  function(x, pivot, ...) {
@@ -544,82 +572,194 @@ diagdiagprod <- function(x, y) {
 		x <- as(x, "dMatrix")
 	    x@x <- as.numeric(nx)
 	}
-	return(x)
+	x
     } else ## x is unit diagonal
-    return(y)
+	y
+}
+
+##' Boolean Algebra/Arithmetic Product of Diagonal Matrices
+##'  %&%
+diagdiagprodBool <- function(x, y) {
+    dimCheck(x,y)
+    if(x@diag != "U") {
+	if(!is.logical(x@x)) x <- as(x, "lMatrix")
+	if(y@diag != "U") {
+	    nx <- x@x & y@x
+	    x@x <- as.logical(nx)
+	}
+	x
+    } else { ## x is unit diagonal: return y
+	if(!is.logical(y@x)) y <- as(y, "lMatrix")
+	y
+    }
 }
 
 setMethod("%*%", signature(x = "diagonalMatrix", y = "diagonalMatrix"),
 	  diagdiagprod, valueClass = "ddiMatrix")
 
-formals(diagdiagprod) <- alist(x=, y=x)
+setMethod("%&%", signature(x = "diagonalMatrix", y = "diagonalMatrix"),
+	  diagdiagprodBool, valueClass = "ldiMatrix")# do *not* have "ndiMatrix" !
+
+##' Both Numeric or Boolean Algebra/Arithmetic Product of Diagonal Matrices
+diagdiagprodFlexi <- function(x, y=NULL, boolArith = NA, ...)
+{
+    dimCheck(x,y)
+    bool <- isTRUE(boolArith)
+    if(x@diag != "U") {
+	if(bool && !is.logical(x@x)) x <- as(x, "lMatrix")
+	if(y@diag != "U") {
+	    if(bool) {
+		nx <- x@x & y@x
+		x@x <- as.logical(nx)
+	    } else { ## boolArith is NA or FALSE: ==> numeric, as have *no* "diagMatrix" patter[n]:
+		nx <- x@x * y@x
+		if(is.numeric(nx) && !is.numeric(x@x))
+		    x <- as(x, "dMatrix")
+		x@x <- as.numeric(nx)
+	    }
+	}
+	x
+    } else { ## x is unit diagonal: return y
+	if(bool && !is.logical(y@x)) y <- as(y, "lMatrix")
+	y
+    }
+}
 setMethod("crossprod", signature(x = "diagonalMatrix", y = "diagonalMatrix"),
-	  diagdiagprod, valueClass = "ddiMatrix")
+	  diagdiagprodFlexi)
 setMethod("tcrossprod", signature(x = "diagonalMatrix", y = "diagonalMatrix"),
-	  diagdiagprod, valueClass = "ddiMatrix")
-setMethod("crossprod", signature(x = "diagonalMatrix", y = "missing"),
-	  diagdiagprod, valueClass = "ddiMatrix")
-setMethod("tcrossprod", signature(x = "diagonalMatrix", y = "missing"),
-	  diagdiagprod, valueClass = "ddiMatrix")
+	  diagdiagprodFlexi)
+
+##' crossprod(x) := x'x
+diagprod <- function(x, y = NULL, boolArith = NA, ...) {
+    bool <- isTRUE(boolArith)
+    if(bool && !is.logical(x@x)) x <- as(x, "lMatrix")
+    if(x@diag != "U") {
+        if(bool) {
+            nx <- x@x & y@x
+            x@x <- as.logical(nx)
+        } else { ## boolArith is NA or FALSE: ==> numeric, as have *no* "diagMatrix" patter[n]:
+            nx <- x@x * x@x
+            if(is.numeric(nx) && !is.numeric(x@x))
+                x <- as(x, "dMatrix")
+            x@x <- as.numeric(nx)
+        }
+    }
+    x
+}
+setMethod( "crossprod", signature(x = "diagonalMatrix", y = "missing"), diagprod)
+setMethod("tcrossprod", signature(x = "diagonalMatrix", y = "missing"), diagprod)
 
 
+## analogous to matdiagprod() below:
 diagmatprod <- function(x, y) {
     ## x is diagonalMatrix
-    dx <- dim(x)
-    dy <- dim(y)
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
-    as(if(x@diag == "U") y else x@x * y, "Matrix")
+    if(x@Dim[2] != nrow(y)) stop("non-matching dimensions")
+    Matrix(if(x@diag == "U") y else x@x * y)
 }
-setMethod("%*%", signature(x = "diagonalMatrix", y = "matrix"),
-	  diagmatprod)
-## sneaky .. :
-formals(diagmatprod) <- alist(x=, y=NULL)
-setMethod("crossprod", signature(x = "diagonalMatrix", y = "matrix"),
-	  diagmatprod)
+setMethod("%*%", signature(x = "diagonalMatrix", y = "matrix"), diagmatprod)
+
+##formals(diagmatprod) <- alist(x=, y=NULL, boolArith = NA, ...=) ## FIXME boolArith
+diagmatprod2 <- function(x, y=NULL, boolArith = NA, ...) {
+    ## x is diagonalMatrix
+    if(x@Dim[2] != nrow(y)) stop("non-matching dimensions")
+    Matrix(if(x@diag == "U") y else x@x * y)
+}
+setMethod("crossprod",  signature(x = "diagonalMatrix", y = "matrix"), diagmatprod2)
 
 diagGeprod <- function(x, y) {
-    dx <- dim(x)
-    dy <- dim(y)
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
+    if(x@Dim[2] != y@Dim[1]) stop("non-matching dimensions")
     if(x@diag != "U")
         y@x <- x@x * y@x
     y
 }
 setMethod("%*%", signature(x= "diagonalMatrix", y= "dgeMatrix"), diagGeprod)
 setMethod("%*%", signature(x= "diagonalMatrix", y= "lgeMatrix"), diagGeprod)
-formals(diagGeprod) <- alist(x=, y=NULL)
-setMethod("crossprod", signature(x = "diagonalMatrix", y = "dgeMatrix"),
-	  diagGeprod, valueClass = "dgeMatrix")
-setMethod("crossprod", signature(x = "diagonalMatrix", y = "lgeMatrix"),
-	  diagGeprod)
 
+diagGeprodBool <- function(x, y) {
+    if(x@Dim[2] != y@Dim[1]) stop("non-matching dimensions")
+    if(!is.logical(y@x)) y <- as(y, "lMatrix")
+    if(x@diag != "U")
+        y@x <- x@x & y@x
+    y
+}
+setMethod("%&%", signature(x= "diagonalMatrix", y= "geMatrix"), diagGeprodBool)
+
+diagGeprod2 <- function(x, y=NULL, boolArith = NA, ...) {
+    if(x@Dim[2] != y@Dim[1]) stop("non-matching dimensions")
+    bool <- isTRUE(boolArith)
+    if(bool && !is.logical(y@x)) y <- as(y, "lMatrix")
+    if(x@diag != "U")
+        y@x <- if(bool) x@x & y@x else x@x * y@x
+    y
+}
+setMethod("crossprod", signature(x = "diagonalMatrix", y = "dgeMatrix"), diagGeprod2)
+setMethod("crossprod", signature(x = "diagonalMatrix", y = "lgeMatrix"), diagGeprod2)
+
+
+## analogous to diagmatprod() above:
 matdiagprod <- function(x, y) {
     dx <- dim(x)
-    dy <- dim(y)
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
+    if(dx[2] != y@Dim[1]) stop("non-matching dimensions")
     Matrix(if(y@diag == "U") x else x * rep(y@x, each = dx[1]))
 }
-setMethod("%*%", signature(x = "matrix", y = "diagonalMatrix"),
-	  matdiagprod)
-formals(matdiagprod) <- alist(x=, y=NULL)
-setMethod("tcrossprod", signature(x = "matrix", y = "diagonalMatrix"),
-	  matdiagprod)
+setMethod("%*%", signature(x = "matrix", y = "diagonalMatrix"), matdiagprod)
 
 gediagprod <- function(x, y) {
     dx <- dim(x)
-    dy <- dim(y)
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
+    if(dx[2] != y@Dim[1]) stop("non-matching dimensions")
     if(y@diag == "N")
 	x@x <- x@x * rep(y@x, each = dx[1])
     x
 }
 setMethod("%*%", signature(x= "dgeMatrix", y= "diagonalMatrix"), gediagprod)
 setMethod("%*%", signature(x= "lgeMatrix", y= "diagonalMatrix"), gediagprod)
-formals(gediagprod) <- alist(x=, y=NULL)
-setMethod("tcrossprod", signature(x = "dgeMatrix", y = "diagonalMatrix"),
-	  gediagprod)
-setMethod("tcrossprod", signature(x = "lgeMatrix", y = "diagonalMatrix"),
-	  gediagprod)
+
+gediagprodBool <- function(x, y) {
+    dx <- dim(x)
+    if(dx[2] != y@Dim[1]) stop("non-matching dimensions")
+    if(!is.logical(x@x)) x <- as(x, "lMatrix")
+    if(y@diag == "N")
+	x@x <- x@x & rep(y@x, each = dx[1])
+    x
+}
+setMethod("%&%", signature(x= "geMatrix", y= "diagonalMatrix"), gediagprodBool)
+
+setMethod("tcrossprod",signature(x = "matrix", y = "diagonalMatrix"),
+          function(x, y=NULL, boolArith = NA, ...) {
+              dx <- dim(x)
+              if(dx[2] != y@Dim[1]) stop("non-matching dimensions")
+              bool <- isTRUE(boolArith)
+              if(bool && !is.logical(y@x)) y <- as(y, "lMatrix")
+              Matrix(if(y@diag == "U") x else
+                     if(bool) x & rep(y@x, each = dx[1])
+                     else     x * rep(y@x, each = dx[1]))
+          })
+
+setMethod("crossprod", signature(x = "matrix", y = "diagonalMatrix"),
+	  function(x, y=NULL, boolArith = NA, ...) {
+	      dx <- dim(x)
+	      if(dx[1] != y@Dim[1]) stop("non-matching dimensions")
+              bool <- isTRUE(boolArith)
+              if(bool && !is.logical(y@x)) y <- as(y, "lMatrix")
+	      Matrix(if(y@diag == "U") t(x) else
+		     if(bool) t(rep.int(y@x, dx[2]) & x)
+		     else     t(rep.int(y@x, dx[2]) * x))
+	  })
+
+
+gediagprod2 <- function(x, y=NULL, boolArith = NA, ...) {
+    dx <- dim(x)
+    if(dx[2] != y@Dim[1]) stop("non-matching dimensions")
+    bool <- isTRUE(boolArith)
+    if(bool && !is.logical(x@x)) x <- as(x, "lMatrix")
+    if(y@diag == "N")
+	x@x <- if(bool) x@x & rep(y@x, each = dx[1])
+	       else     x@x * rep(y@x, each = dx[1])
+    x
+}
+setMethod("tcrossprod", signature(x = "dgeMatrix", y = "diagonalMatrix"), gediagprod2)
+setMethod("tcrossprod", signature(x = "lgeMatrix", y = "diagonalMatrix"), gediagprod2)
+
 
 ## crossprod {more of these}
 
@@ -636,82 +776,146 @@ setMethod("%*%", signature(x = "denseMatrix", y = "diagonalMatrix"),
 ## 	  function(x, y = NULL) {
 ##           })
 
-Cspdiagprod <- function(x, y) {
-    dx <- dim(x <- .Call(Csparse_diagU2N, x))
-    dy <- dim(y)
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
-    if(y@diag == "N") {
-	if(!all(y@x[1L] == y@x[-1L]) && is(x, "symmetricMatrix"))
+##' @param x CsparseMatrix
+##' @param y diagonalMatrix
+##' @return x %*% y
+Cspdiagprod <- function(x, y, boolArith = NA, ...) {
+    if((m <- ncol(x)) != y@Dim[1]) stop("non-matching dimensions")
+    if(y@diag == "N") { ## otherwise: y == Diagonal(n) : multiplication is identity
+	x <- .Call(Csparse_diagU2N, x)
+	cx <- getClass(class(x))
+	if(!all(y@x[1L] == y@x[-1L]) && extends(cx, "symmetricMatrix"))
 	    x <- as(x, "generalMatrix")
-	ind <- rep.int(seq_len(dx[2]), x@p[-1] - x@p[-dx[2]-1L])
-	x@x <- x@x * y@x[ind]
-    }
-    if(is(x, "compMatrix") && length(xf <- x@factors)) {
-        ## instead of dropping all factors, be smart about some
-	## TODO ......
-	x@factors <- list()
-    }
-    x
-}
-
-diagCspprod <- function(x, y) {
-    dx <- dim(x)
-    dy <- dim(y <- .Call(Csparse_diagU2N, y))
-    if(dx[2] != dy[1]) stop("non-matching dimensions")
-    if(x@diag == "N") {
-	if(!all(x@x[1L] == x@x[-1L]) && is(y, "symmetricMatrix"))
-	    y <- as(y, "generalMatrix")
-	y@x <- y@x * x@x[y@i + 1L]
-    }
-    if(is(y, "compMatrix") && length(yf <- y@factors)) {
-	## instead of dropping all factors, be smart about some
-	## TODO
-	keep <- character()
-	if(iLU <- names(yf) == "LU") {
-	    ## TODO keep <- "LU"
+	ind <- rep.int(seq_len(m), x@p[-1] - x@p[-m-1L])
+	if(isTRUE(boolArith)) {
+	    if(extends(cx, "nMatrix")) x <- as(x, "lMatrix") # so, has y@x
+	    x@x <- r <- x@x & y@x[x@i + 1L]
+	    if(!anyNA(r) && !extends(cx, "diagonalMatrix")) x <- as(drop0(x), "nMatrix")
+	} else {
+	    if(!extends(cx, "dMatrix")) x <- as(x, "dMatrix") # <- FIXME if we have zMatrix
+	    x@x <- x@x * y@x[ind]
+	}
+        if(.hasSlot(x, "factors") && length(x@factors)) {# drop cashed ones
+	    ## instead of dropping all factors, be smart about some
+	    ## TODO ......
+	    x@factors <- list()
+	}
+        x
+    } else { #	y is unit-diagonal ==> "return x"
+	cx <- getClass(class(x))
+	if(isTRUE(boolArith)) {
+	    is.l <- if(extends(cx, "dMatrix")) { ## <- FIXME: extend once we have iMatrix, zMatrix
+		x <- as(x, "lMatrix"); TRUE } else extends(cx, "lMatrix")
+	    if(is.l && !anyNA(x@x)) as(drop0(x), "nMatrix")
+	    else if(is.l) x else # defensive:
+	    as(x, "lMatrix")
+	} else {
+	    ## else boolArith is  NA or FALSE {which are equivalent here, das diagonal = "numLike"}
+	    if(extends(cx, "nMatrix") || extends(cx, "lMatrix"))
+		as(x, "dMatrix") else x
 	}
-	y@factors <- yf[keep]
     }
-    y
 }
 
+##' @param x diagonalMatrix
+##' @param y CsparseMatrix
+##' @return x %*% y
+diagCspprod <- function(x, y, boolArith = NA, ...) {
+    if(x@Dim[2] != y@Dim[1]) stop("non-matching dimensions")
+    if(x@diag == "N") {
+	y <- .Call(Csparse_diagU2N, y)
+	cy <- getClass(class(y))
+	if(!all(x@x[1L] == x@x[-1L]) && extends(cy, "symmetricMatrix"))
+	    y <- as(y, "generalMatrix")
+	if(isTRUE(boolArith)) {
+	    if(extends(cy, "nMatrix")) y <- as(y, "lMatrix") # so, has y@x
+	    y@x <- r <- y@x & x@x[y@i + 1L]
+	    if(!anyNA(r) && !extends(cy, "diagonalMatrix")) y <- as(drop0(y), "nMatrix")
+	} else {
+	    if(!extends(cy, "dMatrix")) y <- as(y, "dMatrix") # <- FIXME if we have zMatrix
+	    y@x <- y@x * x@x[y@i + 1L]
+	}
+	if(.hasSlot(y, "factors") && length(y@factors)) {
+     ## if(.hasSlot(y, "factors") && length(yf <- y@factors)) { ## -- TODO? --
+	    ## instead of dropping all factors, be smart about some
+	    ## keep <- character()
+	    ## if(any(names(yf) == "LU")) { ## <- not easy: y = P'LUQ,  x y = xP'LUQ => LU ???
+	    ##     keep <- "LU"
+	    ## }
+	    ## y@factors <- yf[keep]
+	    y@factors <- list()
+        }
+        y
+    } else { ## x @ diag  == "U"
+	cy <- getClass(class(y))
+	if(isTRUE(boolArith)) {
+	    is.l <- if(extends(cy, "dMatrix")) { ## <- FIXME: extend once we have iMatrix, zMatrix
+		y <- as(y, "lMatrix"); TRUE } else extends(cy, "lMatrix")
+	    if(is.l && !anyNA(y@x)) as(drop0(y), "nMatrix")
+	    else if(is.l) y else # defensive:
+	    as(y, "lMatrix")
+	} else {
+	    ## else boolArith is  NA or FALSE {which are equivalent here, das diagonal = "numLike"}
+	    if(extends(cy, "nMatrix") || extends(cy, "lMatrix"))
+		as(y, "dMatrix") else y
+	}
+    }
+}
+
+## + 'boolArith' argument  { ==> .local() is used in any case; keep formals simple :}
 setMethod("crossprod", signature(x = "diagonalMatrix", y = "CsparseMatrix"),
-	  function(x, y = NULL) diagCspprod(x, y))
+	  function(x, y=NULL, boolArith=NA, ...) diagCspprod(x, y, boolArith=boolArith))
 
 setMethod("crossprod", signature(x = "diagonalMatrix", y = "sparseMatrix"),
-	  function(x, y = NULL) diagCspprod(x, as(y, "CsparseMatrix")))
+	  function(x, y=NULL, boolArith=NA, ...)
+	      diagCspprod(x, as(y, "CsparseMatrix"), boolArith=boolArith))
 
 ## Prefer calling diagCspprod to Cspdiagprod if going to transpose anyway
 ##  x'y == (y'x)'
 setMethod("crossprod", signature(x = "CsparseMatrix", y = "diagonalMatrix"),
-	  function(x, y = NULL) t(diagCspprod(y, x)))
+	  function(x, y=NULL, boolArith=NA, ...) t(diagCspprod(y, x, boolArith=boolArith)))
 
 setMethod("crossprod", signature(x = "sparseMatrix", y = "diagonalMatrix"),
-	  function(x, y = NULL) t(diagCspprod(y, as(x, "Csparsematrix"))))
+	  function(x, y=NULL, boolArith=NA, ...) t(diagCspprod(y, as(x, "Csparsematrix"), boolArith=boolArith)))
 
 setMethod("tcrossprod", signature(x = "diagonalMatrix", y = "CsparseMatrix"),
-	  function(x, y = NULL) diagCspprod(x, t(y)))
+	  function(x, y=NULL, boolArith=NA, ...) diagCspprod(x, t(y), boolArith=boolArith))
 
 setMethod("tcrossprod", signature(x = "diagonalMatrix", y = "sparseMatrix"),
-	  function(x, y = NULL) diagCspprod(x, t(as(y, "CsparseMatrix"))))
+	  function(x, y=NULL, boolArith=NA, ...) diagCspprod(x, t(as(y, "CsparseMatrix")), boolArith=boolArith))
 
 setMethod("tcrossprod", signature(x = "CsparseMatrix", y = "diagonalMatrix"),
-	  function(x, y = NULL) Cspdiagprod(x, y))
+	  function(x, y=NULL, boolArith=NA, ...) Cspdiagprod(x, y, boolArith=boolArith))
 
 setMethod("tcrossprod", signature(x = "sparseMatrix", y = "diagonalMatrix"),
-	  function(x, y = NULL) Cspdiagprod(as(x, "CsparseMatrix"), y))
+	  function(x, y=NULL, boolArith=NA, ...) Cspdiagprod(as(x, "CsparseMatrix"), y, boolArith=boolArith))
 
 setMethod("%*%", signature(x = "diagonalMatrix", y = "CsparseMatrix"),
-	  function(x, y) diagCspprod(x, y))
+	  function(x, y) diagCspprod(x, y, boolArith=NA))
+setMethod("%&%", signature(x = "diagonalMatrix", y = "CsparseMatrix"),
+	  function(x, y) diagCspprod(x, y, boolArith=TRUE))
+
+## instead of "sparseMatrix", use: [RT]sparse.. ("closer" in method dispatch)
+for(cl in c("TsparseMatrix", "RsparseMatrix")) {
 
 setMethod("%*%", signature(x = "diagonalMatrix", y = "sparseMatrix"),
-	  function(x, y) diagCspprod(as(x, "CsparseMatrix"), y))
+	  function(x, y) diagCspprod(as(x, "CsparseMatrix"), y, boolArith=NA))
 
 setMethod("%*%", signature(x = "sparseMatrix", y = "diagonalMatrix"),
-	  function(x, y) Cspdiagprod(as(x, "CsparseMatrix"), y))
+	  function(x, y) Cspdiagprod(as(x, "CsparseMatrix"), y, boolArith=NA))
+
+setMethod("%&%", signature(x = "diagonalMatrix", y = "sparseMatrix"),
+	  function(x, y) diagCspprod(as(x, "CsparseMatrix"), y, boolArith=TRUE))
+
+setMethod("%&%", signature(x = "sparseMatrix", y = "diagonalMatrix"),
+	  function(x, y) Cspdiagprod(as(x, "CsparseMatrix"), y, boolArith=TRUE))
+
+}
 
 setMethod("%*%", signature(x = "CsparseMatrix", y = "diagonalMatrix"),
-	  function(x, y) Cspdiagprod(x, y))
+	  function(x, y) Cspdiagprod(x, y, boolArith=NA))
+setMethod("%&%", signature(x = "CsparseMatrix", y = "diagonalMatrix"),
+	  function(x, y) Cspdiagprod(x, y, boolArith=TRUE))
 
 ## TODO: Write tests in ./tests/ which ensure that many "ops" with diagonal*
 ##       do indeed work by going through sparse (and *not* ddense)!
@@ -777,7 +981,7 @@ diagOdiag <- function(e1,e2) {
 	xx <- as.vector(matrix(rbind(r, matrix(r00,n,n)), n,n))
 	newcl <-
 	    paste0(if(isNum) "d" else if(isLog) {
-		if(!any(is.na(r)) && !any(is.na(r00))) "n" else "l"
+		if(!anyNA(r) && !anyNA(r00)) "n" else "l"
 	    } else stop("not yet implemented .. please report"), "syMatrix")
 
 	new(newcl, Dim = e1@Dim, Dimnames = e1@Dimnames, x = xx)
@@ -804,7 +1008,7 @@ diagOtri <- function(e1,e2) {
     ## result must be triangular
     r <- callGeneric(d1 <- .diag.x(e1), diag(e2)) # error if not "compatible"
     ## Check what happens with non-diagonals, i.e. (0 o 0), (FALSE o 0), ...:
-    e1.0 <- if(.n1 <- is.numeric(d1   )) 0 else FALSE
+    e1.0 <- if(is.numeric(d1)) 0 else FALSE
     r00 <- callGeneric(e1.0, if(.n2 <- is.numeric(e2[0])) 0 else FALSE)
     if(is0(r00)) { ##  r00 == 0 or FALSE --- result *is* triangular
         diag(e2) <- r
@@ -889,7 +1093,7 @@ setMethod("Arith", signature(e1 = "ddiMatrix", e2 = arg2),
 	  function(e1,e2) {
 	      n <- e1@Dim[1]
 	      f0 <- callGeneric(0, e2)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e2)) == 1L
 		  if(e1@diag == "U") {
 		      if(any((r <- callGeneric(1, e2)) != 1)) {
@@ -911,7 +1115,7 @@ setMethod("Arith", signature(e1 = arg1, e2 = "ddiMatrix"),
 	  function(e1,e2) {
 	      n <- e2@Dim[1]
 	      f0 <- callGeneric(e1, 0)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e1)) == 1L
 		  if(e2@diag == "U") {
 		      if(any((r <- callGeneric(e1, 1)) != 1)) {
@@ -934,10 +1138,9 @@ setMethod("Arith", signature(e1 = "ldiMatrix", e2 = arg2),
 	  function(e1,e2) {
 	      n <- e1@Dim[1]
 	      f0 <- callGeneric(0, e2)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e2)) == 1L
-		  E <- copyClass(e1, "ddiMatrix", c("diag", "Dim", "Dimnames"))#FIXME: if ok, check=FALSE
-		  ## E <- copyClass(e1, "ddiMatrix", check=FALSE)
+		  E <- copyClass(e1, "ddiMatrix", c("diag", "Dim", "Dimnames"), check=FALSE)
 		  ## storage.mode(E@x) <- "double"
 		  if(e1@diag == "U") {
 		      if(any((r <- callGeneric(1, e2)) != 1)) {
@@ -959,10 +1162,9 @@ setMethod("Arith", signature(e1 = arg1, e2 = "ldiMatrix"),
 	  function(e1,e2) {
 	      n <- e2@Dim[1]
 	      f0 <- callGeneric(e1, 0)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e1)) == 1L
-		  E <- copyClass(e2, "ddiMatrix", c("diag", "Dim", "Dimnames"))#FIXME: if ok, check=FALSE
-		  ## E <- copyClass(e2, "ddiMatrix", check=FALSE)
+		  E <- copyClass(e2, "ddiMatrix", c("diag", "Dim", "Dimnames"), check=FALSE)
 		  ## storage.mode(E@x) <- "double"
 		  if(e2@diag == "U") {
 		      if(any((r <- callGeneric(e1, 1)) != 1)) {
@@ -992,10 +1194,9 @@ setMethod("Ops", signature(e1 = "ddiMatrix", e2 = arg2),
 	  function(e1,e2) {
 	      n <- e1@Dim[1]
 	      f0 <- callGeneric(0, e2)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e2)) == 1L
-		  E <- copyClass(e1, "ldiMatrix", c("diag", "Dim", "Dimnames"))#FIXME: if ok, check=FALSE
-		  ## E <- copyClass(e1, "ldiMatrix", check=FALSE)
+		  E <- copyClass(e1, "ldiMatrix", c("diag", "Dim", "Dimnames"), check=FALSE)
 		  ## storage.mode(E@x) <- "logical"
 		  if(e1@diag == "U") {
 		      if(any((r <- callGeneric(1, e2)) != 1)) {
@@ -1018,7 +1219,7 @@ setMethod("Ops", signature(e1 = "ldiMatrix", e2 = arg2),
 	  function(e1,e2) {
 	      n <- e1@Dim[1]
 	      f0 <- callGeneric(FALSE, e2)
-	      if(all(is0(f0))) { # remain diagonal
+	      if(all0(f0)) { # remain diagonal
 		  L1 <- (le <- length(e2)) == 1L
 
 		  if(e1@diag == "U") {
@@ -1052,8 +1253,10 @@ for(other in c("ANY", "Matrix", "dMatrix")) {
 }
 
 ## Direct subclasses of "denseMatrix": currently ddenseMatrix, ldense... :
+if(FALSE) # now also contains "geMatrix"
 dense.subCl <- local({ dM.scl <- getClass("denseMatrix")@subclasses
 		       names(dM.scl)[vapply(dM.scl, slot, 0, "distance") == 1] })
+dense.subCl <- paste0(c("d","l","n"), "denseMatrix")
 for(DI in diCls) {
     dMeth <- if(extends(DI, "dMatrix"))
 	function(e1,e2) callGeneric(diag2Tsmart(e1,e2, "d"), e2)
@@ -1073,6 +1276,7 @@ for(DI in diCls) {
     }
 }
 
+## Group methods "Math", "Math2" in			--> ./Math.R
 
 ### "Summary" : "max"   "min"   "range" "prod"  "sum"   "any"   "all"
 ### ----------   the last 4: separately here
@@ -1128,7 +1332,7 @@ prDiag <-
 {
     cf <- array(".", dim = x@Dim, dimnames = x@Dimnames)
     cf[row(cf) == col(cf)] <-
-        sapply(diag(x), format, digits = digits, justify = justify)
+	vapply(diag(x), format, "", digits = digits, justify = justify)
     print(cf, quote = FALSE, right = right)
     invisible(x)
 }
@@ -1152,7 +1356,8 @@ setMethod("show", signature(object = "diagonalMatrix"),
 	      }
 	  })
 
-rm(dense.subCl, diCls)# not used elsewhere
+rm(arg1, arg2, other, DI, Fun, cl, c1, c2,
+   dense.subCl, diCls)# not used elsewhere
 
 setMethod("summary", signature(object = "diagonalMatrix"),
 	  function(object, ...) {
