@@ -7,6 +7,13 @@
 
 setAs("ANY", "sparseMatrix", function(from) as(from, "CsparseMatrix"))
 
+## If people did not use xtabs(), but table():
+setAs("table", "sparseMatrix", function(from) {
+    if(length(dim(from)) != 2L)
+        stop("only 2-dimensional tables can be directly coerced to sparse matrices")
+    as(unclass(from), "CsparseMatrix")
+})
+
 setAs("sparseMatrix", "generalMatrix", as_gSparse)
 
 setAs("sparseMatrix", "symmetricMatrix", as_sSparse)
@@ -20,14 +27,16 @@ spMatrix <- function(nrow, ncol,
     ## The conformability of (i,j,x) with itself and with 'dim'
     ## is checked automatically by internal "validObject()" inside new(.):
     kind <- .M.kind(x)
-    new(paste(kind, "gTMatrix", sep=''), Dim = dim,
+    new(paste0(kind, "gTMatrix"), Dim = dim,
         x = if(kind == "d") as.double(x) else x,
         ## our "Tsparse" Matrices use  0-based indices :
         i = as.integer(i - 1L),
         j = as.integer(j - 1L))
 }
 
-sparseMatrix <- function(i = ep, j = ep, p, x, dims, dimnames, index1 = TRUE)
+sparseMatrix <- function(i = ep, j = ep, p, x, dims, dimnames,
+                         symmetric = FALSE, index1 = TRUE,
+                         giveCsparse = TRUE, check = TRUE, use.last.ij = FALSE)
 {
   ## Purpose: user-level substitute for most  new(<sparseMatrix>, ..) calls
   ## Author: Douglas Bates, Date: 12 Jan 2009, based on Martin's version
@@ -39,39 +48,56 @@ sparseMatrix <- function(i = ep, j = ep, p, x, dims, dimnames, index1 = TRUE)
             stop("'p' must be a non-decreasing vector (0, ...)")
         ep <- rep.int(seq_along(dp), dp)
     }
-    ## i and j are now both defined.  Make them 1-based indices.
+    ## i and j are now both defined (via default = ep).  Make them 1-based indices.
     i1 <- as.logical(index1)[1]
     i <- as.integer(i + !(m.i || i1))
     j <- as.integer(j + !(m.j || i1))
 
     ## "minimal dimensions" from (i,j,p); no warnings from empty i or j :
     dims.min <- suppressWarnings(c(max(i), max(j)))
-    if(any(is.na(dims.min))) stop("NA's in (i,j) are not allowed")
+    if(anyNA(dims.min)) stop("NA's in (i,j) are not allowed")
     if(missing(dims)) {
         dims <- dims.min
     } else { ## check dims
         stopifnot(all(dims >= dims.min))
         dims <- as.integer(dims)
     }
+    sx <- if(symmetric) {
+        if(dims[1] != dims[2])
+            stop("symmetric matrix must be square")
+        "s"
+    } else "g"
     isPat <- missing(x) ## <-> patter"n" Matrix
     kx <- if(isPat) "n" else .M.kind(x)
-    r <- new(paste(kx, "gTMatrix", sep=''))
+    r <- new(paste0(kx, sx, "TMatrix"))
     r@Dim <- dims
+    if(symmetric && all(i >= j)) r@uplo <- "L" # else "U", the default
     if(!isPat) {
 	if(kx == "d" && !is.double(x)) x <- as.double(x)
 	if(length(x) != (n <- length(i))) { ## recycle
 	    if(length(x) != 1 && n %% length(x) != 0)
 		warning("length(i) is not a multiple of length(x)")
-	    x <- rep(x, length.out = n)
+	    x <- rep_len(x, n)
 	}
+        if(use.last.ij && (id <- anyDuplicated(cbind(i,j), fromLast=TRUE))) {
+            i <- i[-id]
+            j <- j[-id]
+            x <- x[-id]
+            if(any(idup <- duplicated(cbind(i,j), fromLast=TRUE))) {
+                ndup <- -which(idup)
+                i <- i[ndup]
+                j <- j[ndup]
+                x <- x[ndup]
+            }
+        }
 	r@x <- x
     }
     r@i <- i - 1L
     r@j <- j - 1L
     if(!missing(dimnames))
-	r@Dimnames <- dimnames
-    validObject(r)
-    as(r, "CsparseMatrix")
+	r@Dimnames <- .fixupDimnames(dimnames)
+    if(check) validObject(r)
+    if(giveCsparse) as(r, "CsparseMatrix") else r
 }
 
 ## "graph" coercions -- this needs the graph package which is currently
@@ -83,6 +109,8 @@ sparseMatrix <- function(i = ep, j = ep, p, x, dims, dimnames, index1 = TRUE)
 ## Add some utils that may no longer be needed in future versions of the 'graph' package
 graph.has.weights <- function(g) "weight" %in% names(graph::edgeDataDefaults(g))
 
+graph.non.1.weights <- function(g) any(unlist(graph::edgeData(g, attr = "weight")) != 1)
+
 graph.wgtMatrix <- function(g)
 {
     ## Purpose: work around "graph" package's  as(g, "matrix") bug
@@ -99,9 +127,11 @@ graph.wgtMatrix <- function(g)
 
     ## Usual case, when there are edges:
     if(has.w <- graph.has.weights(g)) {
+        ## graph.non.1.weights(g) :
         w <- unlist(graph::edgeData(g, attr = "weight"))
         has.w <- any(w != 1)
     } ## now 'has.w' is TRUE  iff  there are weights != 1
+    ## now 'has.w' is TRUE  iff  there are weights != 1
     m <- as(g, "matrix")
     ## now is a 0/1 - matrix (instead of 0/wgts) with the 'graph' bug
     if(has.w) { ## fix it if needed
@@ -134,50 +164,54 @@ setAs("graph", "Matrix", function(from) as(from, "CsparseMatrix"))
 setAs("graphNEL", "CsparseMatrix",
       function(from) as(as(from, "TsparseMatrix"), "CsparseMatrix"))
 
-setAs("graphNEL", "TsparseMatrix",
-      function(from) {
-          nd <- graph::nodes(from)
-          dm <- rep.int(length(nd), 2)
-	  symm <- graph::edgemode(from) == "undirected"
+graph2T <- function(from, use.weights =
+		    graph.has.weights(from) && graph.non.1.weights(from)) {
+    nd <- graph::nodes(from); dnms <- list(nd,nd)
+    dm <- rep.int(length(nd), 2)
+    edge2i <- function(e) {
+	## return (0-based) row indices 'i'
+	rep.int(0:(dm[1]-1L), lengths(e))
+    }
 
- 	  if(graph.has.weights(from)) {
-	      eWts <- graph::edgeWeights(from)
-	      lens <- unlist(lapply(eWts, length))
-	      i <- rep.int(0:(dm[1]-1), lens) # column indices (0-based)
-	      To <- unlist(lapply(eWts, names))
-	      j <- as.integer(match(To,nd) - 1L) # row indices (0-based)
-	      ## symm <- symm && <weights must also be symmetric>: improbable
-	      ## if(symm) new("dsTMatrix", .....) else
-	      new("dgTMatrix", i = i, j = j, x = unlist(eWts),
-		  Dim = dm, Dimnames = list(nd, nd))
-	  }
- 	  else { ## no weights: 0/1 matrix -> logical
-              edges <- lapply(from@edgeL[nd], "[[", "edges")
-              lens <- unlist(lapply(edges, length))
-              ## nnz <- sum(unlist(lens))  # number of non-zeros
-              i <- rep.int(0:(dm[1]-1), lens) # column indices (0-based)
-              j <- as.integer(unlist(edges) - 1) # row indices (0-based)
-              if(symm) {            # symmetric: ensure upper triangle
-                  tmp <- i
-                  flip <- i > j
-                  i[flip] <- j[flip]
-                  j[flip] <- tmp[flip]
-                  new("nsTMatrix", i = i, j = j, Dim = dm,
-                      Dimnames = list(nd, nd), uplo = "U")
-              } else {
-                  new("ngTMatrix", i = i, j = j, Dim = dm,
-                      Dimnames = list(nd, nd))
-              }
-          }
-      })
+    if(use.weights) {
+	eWts <- graph::edgeWeights(from); names(eWts) <- NULL
+	i <- edge2i(eWts)
+	To <- unlist(lapply(eWts, names))
+	j <- as.integer(match(To,nd)) - 1L # columns indices (0-based)
+	## symm <- symm && <weights must also be symmetric>: improbable
+	## if(symm) new("dsTMatrix", .....) else
+	new("dgTMatrix", i = i, j = j, x = unlist(eWts), Dim = dm, Dimnames = dnms)
+    }
+    else { ## no weights: 0/1 matrix -> logical
+	edges <- lapply(from@edgeL[nd], "[[", "edges")
+	symm <- graph::edgemode(from) == "undirected"
+	if(symm)# each edge appears twice; keep upper triangle only
+	    edges <- lapply(seq_along(edges), function(i) {e <- edges[[i]]; e[e >= i]})
+	i <- edge2i(edges)
+	j <- as.integer(unlist(edges)) - 1L # column indices (0-based)
+	## if(symm) {			# symmetric: ensure upper triangle
+	##     tmp <- i
+	##     flip <- i > j
+	##     i[flip] <- j[flip]
+	##     j[flip] <- tmp[flip]
+	##     new("nsTMatrix", i = i, j = j, Dim = dm, Dimnames = dnms, uplo = "U")
+	## } else {
+	##     new("ngTMatrix", i = i, j = j, Dim = dm, Dimnames = dnms)
+	## }
+	new(if(symm) "nsTMatrix" else "ngTMatrix",
+	    i = i, j = j, Dim = dm, Dimnames = dnms)# uplo = "U" is default
+    }
+}
+setAs("graphNEL", "TsparseMatrix", function(from) graph2T(from))
 
 setAs("sparseMatrix", "graph", function(from) as(from, "graphNEL"))
 setAs("sparseMatrix", "graphNEL",
-      ## since have specific method for Tsparse below, are Csparse or Rsparse,
+      ## since have specific method for Tsparse below, 'from' is *not*,
       ## i.e. do not need to "uniquify" the T* matrix:
-      function(from) Tsp2grNEL(as(from, "TsparseMatrix"), need.uniq=FALSE))
+      function(from) T2graph(as(from, "TsparseMatrix"), need.uniq=FALSE))
+setAs("TsparseMatrix", "graphNEL", function(from) T2graph(from))
 
-Tsp2grNEL <- function(from, need.uniq = is_not_uniqT(from)) {
+T2graph <- function(from, need.uniq = is_not_uniqT(from), edgemode = NULL) {
     d <- dim(from)
     if(d[1] != d[2])
 	stop("only square matrices can be used as incidence matrices for graphs")
@@ -186,24 +220,26 @@ Tsp2grNEL <- function(from, need.uniq = is_not_uniqT(from)) {
     if(is.null(rn <- dimnames(from)[[1]]))
 	rn <- as.character(1:n)
     if(need.uniq) ## Need to 'uniquify' the triplets!
-        from <- uniq(from)
+	from <- uniqTsparse(from)
 
-    if(isSymmetric(from)) { # either "symmetricMatrix" or otherwise
-	##-> undirected graph: every edge only once!
-	if(!is(from, "symmetricMatrix")) {
-	    ## a general matrix which happens to be symmetric
-	    ## ==> remove the double indices
-	    from <- tril(from)
-	}
-        eMode <- "undirected"
-    } else {
-        eMode <- "directed"
-    }
+    if(is.null(edgemode))
+        edgemode <-
+            if(isSymmetric(from)) { # either "symmetricMatrix" or otherwise
+                ##-> undirected graph: every edge only once!
+                if(!is(from, "symmetricMatrix")) {
+                    ## a general matrix which happens to be symmetric
+                    ## ==> remove the double indices
+                    from <- tril(from)
+                }
+                "undirected"
+            } else {
+                "directed"
+            }
     ## every edge is there only once, either upper or lower triangle
     ft1 <- cbind(rn[from@i + 1L], rn[from@j + 1L])
-    graph::ftM2graphNEL(ft1, W = from@x, V= rn, edgemode= eMode)
+    graph::ftM2graphNEL(ft1, W = if(.hasSlot(from,"x")) as.numeric(from@x), ## else NULL
+			V = rn, edgemode=edgemode)
 }
-setAs("TsparseMatrix", "graphNEL", function(from) Tsp2grNEL(from))
 
 
 ### Subsetting -- basic things (drop = "missing") are done in ./Matrix.R
@@ -236,7 +272,7 @@ setMethod("[", signature(x = "sparseMatrix", i = "missing", j = "index",
 	      cld <- getClassDef(class(x))
 ##> why should this be needed; can still happen in <Tsparse>[..]:
 ##>	      if(!extends(cld, "generalMatrix")) x <- as(x, "generalMatrix")
-##	      viaCl <- paste(.M.kind(x, cld), "gTMatrix", sep='')
+##	      viaCl <- paste0(.M.kind(x, cld), "gTMatrix")
 
 	      x <- as(x, "TsparseMatrix")[, j, drop=drop]
 ##simpler than x <- callGeneric(x = as(x, "TsparseMatrix"), j=j, drop=drop)
@@ -255,8 +291,8 @@ setMethod("[", signature(x = "sparseMatrix",
 ##> why should this be needed; can still happen in <Tsparse>[..]:
 ##>	      if(!doSym && !extends(cld, "generalMatrix"))
 ##>		  x <- as(x, "generalMatrix")
-##	      viaCl <- paste(.M.kind(x, cld),
-##			     if(doSym) "sTMatrix" else "gTMatrix", sep='')
+##	      viaCl <- paste0(.M.kind(x, cld),
+##			      if(doSym) "sTMatrix" else "gTMatrix")
 	      x <- as(x, "TsparseMatrix")[i, j, drop=drop]
 	      if(is(x, "Matrix") && extends(cld, "CsparseMatrix"))
 		  as(x, "CsparseMatrix") else x
@@ -271,9 +307,10 @@ setMethod("[", signature(x = "sparseMatrix",
 ## x[] <- value :
 setReplaceMethod("[", signature(x = "sparseMatrix", i = "missing", j = "missing",
 				value = "ANY"),## double/logical/...
-	  function (x, value) {
+	  function (x, i,j,..., value) {
 	      if(all0(value)) { # be faster
 		  cld <- getClassDef(class(x))
+		  x <- diagU2N(x, cl = cld)
 		  for(nm in intersect(nsl <- names(cld@slots),
 				      c("x", "i","j", "factors")))
 		      length(slot(x, nm)) <- 0L
@@ -289,12 +326,15 @@ setReplaceMethod("[", signature(x = "sparseMatrix", i = "missing", j = "missing"
 setReplaceMethod("[", signature(x = "sparseMatrix", i = "missing", j = "ANY",
 				value = "sparseMatrix"),
 		 function (x, i, j, ..., value)
-		 callGeneric(x=x, , j=j, value = as(value, "sparseVector")))
+		     callGeneric(x=x, , j=j, value = as(value, "sparseVector")))
 
 setReplaceMethod("[", signature(x = "sparseMatrix", i = "ANY", j = "missing",
 				value = "sparseMatrix"),
 		 function (x, i, j, ..., value)
-		 callGeneric(x=x, i=i, , value = as(value, "sparseVector")))
+		     if(nargs() == 3)
+			 callGeneric(x=x, i=i, value = as(value, "sparseVector"))
+		     else
+			 callGeneric(x=x, i=i, , value = as(value, "sparseVector")))
 
 setReplaceMethod("[", signature(x = "sparseMatrix", i = "ANY", j = "ANY",
 				value = "sparseMatrix"),
@@ -303,16 +343,6 @@ setReplaceMethod("[", signature(x = "sparseMatrix", i = "ANY", j = "ANY",
 
 
 
-## Group Methods
-
-setMethod("Math",
-	  signature(x = "sparseMatrix"),
-	  function(x) callGeneric(as(x, "CsparseMatrix")))
-
-## further group methods -> see ./Ops.R {"Summary": ./dMatrix.R }
-
-
-
 ### --- print() and show() methods ---
 
 .formatSparseSimple <- function(m, asLogical=FALSE, digits=NULL,
@@ -327,7 +357,13 @@ setMethod("Math",
 	if(is.null(dim(cx))) {# e.g. in	1 x 1 case
 	    dim(cx) <- dim(m)
 	    dimnames(cx) <- dn
-	}
+	} else ## workaround bug in apply() which has lost row names:
+	    if(getRversion() < "3.2" && !is.null(names(dn))) {
+		if(is.null(dimnames(cx)))
+		    dimnames(cx) <- dn
+		else
+		    names(dimnames(cx)) <- names(dn)
+            }
     }
     if (missing(col.names))
 	col.names <- {
@@ -353,7 +389,7 @@ setMethod("Math",
 		   iarg <- as.integer(sub("^[^0-9]*", '', cn))
 		   colnames(cx) <- substr(colnames(cx), 1, iarg)
 	       },
-	       stop("invalid 'col.names' string: ", cn))
+	       stop(gettextf("invalid 'col.names' string: %s", cn), domain=NA))
     }
     ## else: nothing to do for col.names == TRUE
     cx
@@ -362,12 +398,12 @@ setMethod("Math",
 
 ### NB: Want this to work also for logical or numeric traditional matrix 'x':
 formatSparseM <- function(x, zero.print = ".", align = c("fancy", "right"),
-                          m = as(x,"matrix"), asLogical=NULL, digits=NULL,
-                          cx, iN0, dn = dimnames(m))
+                          m = as(x,"matrix"), asLogical=NULL, uniDiag=NULL,
+                          digits=NULL, cx, iN0, dn = dimnames(m))
 {
     cld <- getClassDef(class(x))
     if(is.null(asLogical)) {
-        binary <- extends(cld,"nsparseMatrix") || extends(cld, "pMatrix")# -> simple T / F
+        binary <- extends(cld,"nsparseMatrix") || extends(cld, "indMatrix")# -> simple T / F
         asLogical <- { binary || extends(cld,"lsparseMatrix") ||
                        extends(cld,"matrix") && is.logical(x) }
 					# has NA and (non-)structural FALSE
@@ -379,9 +415,8 @@ formatSparseM <- function(x, zero.print = ".", align = c("fancy", "right"),
 	dimnames(cx) <- dn
     }
     if(missing(iN0))
-	iN0 <- 1L + .Call(m_encodeInd, non0ind(x, cld), di = d, FALSE)
+	iN0 <- 1L + .Call(m_encodeInd, non0ind(x, cld), di = d, FALSE, FALSE)
     ## ne <- length(iN0)
-
     if(asLogical) {
         cx[m] <- "|"
         if(!extends(cld, "sparseMatrix"))
@@ -402,7 +437,7 @@ formatSparseM <- function(x, zero.print = ".", align = c("fancy", "right"),
                 ij <- rbind(ij, ij[notdiag, 2:1], deparse.level=0)
                 F. <-	  c(F., F.[notdiag])
             }
-            iN0 <- 1L + .Call(m_encodeInd, ij, di = d, FALSE)
+            iN0 <- 1L + .Call(m_encodeInd, ij, di = d, FALSE, FALSE)
             cx[iN0[F.]] <- ":" # non-structural FALSE (or "o", "," , "-" or "f")?
         }
     }
@@ -430,6 +465,20 @@ formatSparseM <- function(x, zero.print = ".", align = c("fancy", "right"),
         else
             zero.print <- rep.int(zero.print, length(cols))
     } ## else "right" : nothing to do
+    if(!asLogical && isTRUE(uniDiag)) { ## use "I" in diagonal -- pad correctly
+        if(any(diag(x) != 1))
+            stop("uniDiag=TRUE, but not all diagonal entries are 1")
+        D <- diag(cx) # use
+        if(any((ir <- regexpr("1", D)) < 0)) {
+            warning("uniDiag=TRUE, not all entries in diagonal coded as 1")
+        } else {
+            ir <- as.vector(ir)
+            nD <- nchar(D, "bytes")
+            ## replace "1..." by "I  " (I plus blanks)
+            substr(D, ir, nD) <- sprintf("I%*s", nD - ir, "")
+            diag(cx) <- D
+        }
+    }
     cx[-iN0] <- zero.print
     cx
 }## formatSparseM()
@@ -439,13 +488,15 @@ formatSparseM <- function(x, zero.print = ".", align = c("fancy", "right"),
 formatSpMatrix <- function(x, digits = NULL, # getOption("digits"),
                            maxp = 1e9, # ~ 1/2 * .Machine$integer.max, ## getOption("max.print"),
                            cld = getClassDef(class(x)), zero.print = ".",
-                           col.names, note.dropping.colnames = TRUE,
+                           col.names, note.dropping.colnames = TRUE, uniDiag = TRUE,
                            align = c("fancy", "right"))
 {
     stopifnot(extends(cld, "sparseMatrix"))
     validObject(x) # have seen seg.faults for invalid objects
     d <- dim(x)
-    if(extends(cld, "triangularMatrix") && x@diag == "U") {
+    unitD <- extends(cld, "triangularMatrix") && x@diag == "U"
+    ## Will note it is *unit*-diagonal by using "I" instead of "1"
+    if(unitD) {
 	if(extends(cld, "CsparseMatrix"))
 	    x <- .Call(Csparse_diagU2N, x)
 	else if(extends(cld, "TsparseMatrix"))
@@ -453,11 +504,10 @@ formatSpMatrix <- function(x, digits = NULL, # getOption("digits"),
 	else {
 	    kind <- .M.kind(x, cld)
 	    x <- .Call(Tsparse_diagU2N,
-		       as(as(x, paste(kind, "Matrix", sep='')), "TsparseMatrix"))
+		       as(as(x, paste0(kind, "Matrix")), "TsparseMatrix"))
 	    cld <- getClassDef(class(x))
 	}
     }
-    ## TODO?  Could note it is *unit*-diagonal, e.g., by using "I" instead of "1" ?
 
     if(prod(d) > maxp) { # "Large" => will be "cut"
         ## only coerce to dense that part which won't be cut :
@@ -467,7 +517,7 @@ formatSpMatrix <- function(x, digits = NULL, # getOption("digits"),
         m <- as(x, "matrix")
     }
     dn <- dimnames(m) ## will be === dimnames(cx)
-    binary <- extends(cld,"nsparseMatrix") || extends(cld, "pMatrix") # -> simple T / F
+    binary <- extends(cld,"nsparseMatrix") || extends(cld, "indMatrix") # -> simple T / F
     logi <- binary || extends(cld,"lsparseMatrix") # has NA and (non-)structural FALSE
     cx <- .formatSparseSimple(m, asLogical = logi, digits=digits,
                               col.names=col.names,
@@ -482,9 +532,10 @@ formatSpMatrix <- function(x, digits = NULL, # getOption("digits"),
 	## -> cannot use 'm' alone
         d <- dim(cx)
 	ne <- length(iN0 <- 1L + .Call(m_encodeInd, non0ind(x, cld),
-				       di = d, FALSE))
+				       di = d, FALSE, FALSE))
 	if(0 < ne && (logi || ne < prod(d))) {
-	    cx <- formatSparseM(x, zero.print, align, m=m, asLogical=logi,
+	    cx <- formatSparseM(x, zero.print, align, m=m,
+				asLogical = logi, uniDiag = unitD & uniDiag,
 				digits=digits, cx=cx, iN0=iN0, dn=dn)
 	} else if (ne == 0)# all zeroes
 	    cx[] <- zero.print
@@ -500,28 +551,29 @@ formatSpMatrix <- function(x, digits = NULL, # getOption("digits"),
 printSpMatrix <- function(x, digits = NULL, # getOption("digits"),
 			  maxp = getOption("max.print"),
 			  cld = getClassDef(class(x)), zero.print = ".",
-			  col.names, note.dropping.colnames = TRUE,
+			  col.names, note.dropping.colnames = TRUE, uniDiag = TRUE,
 			  col.trailer = '', align = c("fancy", "right"))
 {
     stopifnot(extends(cld, "sparseMatrix"))
-    x.orig <- x # to be returned
     cx <- formatSpMatrix(x, digits=digits, maxp=maxp, cld=cld,
 			 zero.print=zero.print, col.names=col.names,
 			 note.dropping.colnames=note.dropping.colnames,
-			 align=align)
+                         uniDiag=uniDiag, align=align)
     if(col.trailer != '')
         cx <- cbind(cx, col.trailer, deparse.level = 0)
     ## right = TRUE : cheap attempt to get better "." alignment
     print(cx, quote = FALSE, right = TRUE, max = maxp)
-    invisible(x.orig)
+    invisible(x)
 } ## printSpMatrix()
 
+##' The "real" show() / print() method, calling the above printSpMatrix():
 printSpMatrix2 <- function(x, digits = NULL, # getOption("digits"),
                            maxp = getOption("max.print"), zero.print = ".",
-                           col.names, note.dropping.colnames = TRUE,
+                           col.names, note.dropping.colnames = TRUE, uniDiag = TRUE,
                            suppRows = NULL, suppCols = NULL,
                            col.trailer = if(suppCols) "......" else "",
-                           align = c("fancy", "right"))
+                           align = c("fancy", "right"),
+                           width = getOption("width"), fitWidth = TRUE)
 {
     d <- dim(x)
     cl <- class(x)
@@ -530,55 +582,79 @@ printSpMatrix2 <- function(x, digits = NULL, # getOption("digits"),
 	" (unitriangular)" else ""
     cat(sprintf('%d x %d sparse Matrix of class "%s"%s\n',
                 d[1], d[2], cl, xtra))
+    setW <-  !missing(width) && width > getOption("width")
+    if(setW) {
+	op <- options(width = width) ; on.exit( options(op) ) }
     if((identical(suppRows,FALSE) && identical(suppCols, FALSE)) ||
        (!isTRUE(suppRows) && !isTRUE(suppCols) && prod(d) <= maxp))
-    {
+    { ## "small matrix" and supp* not TRUE : no rows or columns are suppressed
         if(missing(col.trailer) && is.null(suppCols))
             suppCols <- FALSE # for 'col.trailer'
         printSpMatrix(x, cld=cld, digits=digits, maxp=maxp,
                       zero.print=zero.print, col.names=col.names,
-                      note.dropping.colnames=note.dropping.colnames,
+                      note.dropping.colnames=note.dropping.colnames, uniDiag=uniDiag,
                       col.trailer=col.trailer, align=align)
-    } else { ## d[1] > maxp / d[2] >= nr : -- this needs [,] working:
+    }
+    else { ## d[1] > maxp / d[2] >= nr : -- this needs [,] working:
 	validObject(x)
-	nR <- d[1] ## nrow
-	useW <- getOption("width") - (format.info(nR, digits=digits)[1] + 3+1)
-	##			     space for "[<last>,] "
-
-	## --> suppress rows and/or columns in printing ...
-
-	if(is.null(suppCols)) suppCols <- (d[2] * 2 > useW)
-	nc <- if(suppCols) (useW - (1 + nchar(col.trailer))) %/% 2 else d[2]
-	nr <- maxp %/% nc
-	if(is.null(suppRows)) suppRows <- (nr < nR)
-
-	sTxt <- c("in show(); maybe adjust 'options(max.print= *)'",
+	sTxt <- c(" ", gettext(
+			   "in show(); maybe adjust 'options(max.print= *, width = *)'"),
 		  "\n ..............................\n")
+	useW <- width - (format.info(d[1], digits=digits)[1] + 3+1)
+	##  ==  width - space for the largest row label : "[<last>,] "
+
+	## Suppress rows and/or columns in printing ...
+        ## ---------------------------------------- but which exactly depends on format
+        ## Determining number of columns - first assuming all zeros : ". . "..: 2 chars/column
+        ## i.e., we get the *maximal* numbers of columns to keep, nc :
+	if(is.null(suppCols)) # i.e., "it depends" ..
+            suppCols <- (d[2] * 2 > useW) # used in 'col.trailer' default
+        nCc <- 1 + nchar(col.trailer, "width")
+	if(suppCols) {
+            nc <- (useW - nCc) %/% 2
+            x <- x[ , 1:nc, drop = FALSE]
+        } else
+            nc <- d[2]
+	nr <- maxp %/% nc # if nc becomes smaller,  nr will become larger (!)
+        if(is.null(suppRows)) suppRows <- (nr < d[1])
 	if(suppRows) {
-	    if(suppCols)
-                x <- x[ , 1:nc, drop = FALSE]
 	    n2 <- ceiling(nr / 2)
-	    printSpMatrix(x[seq_len(min(nR, max(1, n2))), , drop=FALSE],
+	    if(fitWidth) {
+		## one iteration of improving the width, by "fake printing" :
+		cM <- formatSpMatrix(x[seq_len(min(d[1], max(1, n2))), , drop = FALSE],
+				     digits=digits, maxp=maxp, zero.print=zero.print,
+				     col.names=col.names, align=align,
+				     note.dropping.colnames=note.dropping.colnames, uniDiag=FALSE)
+                ## width needed (without the 'col.trailer's  'nCc'):
+		matW <- nchar(capture.output(print(cM, quote=FALSE, right=FALSE))[[1]])
+                needW <- matW + (if(suppCols) nCc else 0)
+                if(needW > useW) { ## need more width
+                    op <- options(width = width+(needW-useW))
+                    if(!setW) on.exit( options(op) )
+                }
+	    }
+	    printSpMatrix(x[seq_len(min(d[1], max(1, n2))), , drop=FALSE],
 			  digits=digits, maxp=maxp,
 			  zero.print=zero.print, col.names=col.names,
-			  note.dropping.colnames=note.dropping.colnames,
+			  note.dropping.colnames=note.dropping.colnames, uniDiag=uniDiag,
 			  col.trailer = col.trailer, align=align)
+	    suppTxt <- gettext(if(suppCols) "suppressing columns and rows" else "suppressing rows")
 	    cat("\n ..............................",
-		"\n ........suppressing rows ", sTxt, "\n", sep='')
+		"\n ........", suppTxt, sTxt, "\n", sep='')
 	    ## tail() automagically uses "[..,]" rownames:
 	    printSpMatrix(tail(x, max(1, nr-n2)),
 			  digits=digits, maxp=maxp,
 			  zero.print=zero.print, col.names=col.names,
-			  note.dropping.colnames=note.dropping.colnames,
+			  note.dropping.colnames=note.dropping.colnames, uniDiag=FALSE,
 			  col.trailer = col.trailer, align=align)
 	}
 	else if(suppCols) {
 	    printSpMatrix(x[ , 1:nc , drop = FALSE],
 			  digits=digits, maxp=maxp,
 			  zero.print=zero.print, col.names=col.names,
-			  note.dropping.colnames=note.dropping.colnames,
+			  note.dropping.colnames=note.dropping.colnames, uniDiag=uniDiag,
 			  col.trailer = col.trailer, align=align)
-	    cat("\n .....suppressing columns ", sTxt, sep='')
+	    cat("\n .....", gettext("suppressing columns"), sTxt, sep='')
 	}
 	else stop("logic programming error in printSpMatrix2(), please report")
 
@@ -620,6 +696,16 @@ print.sparseSummary <- function (x, ...) {
     invisible(x)
 }
 
+
+
+### FIXME [from ../TODO ]: Use cholmod_symmetry() --
+## Possibly even use 'option' as argument here for fast check to use sparse solve !!
+
+##' This case should be particularly fast
+setMethod("isSymmetric", signature(object = "dgCMatrix"),
+	  function(object, tol = 100*.Machine$double.eps, ...)
+	      isTRUE(all.equal(.dgC.0.factors(object), t(object), tolerance = tol, ...)))
+
 setMethod("isSymmetric", signature(object = "sparseMatrix"),
 	  function(object, tol = 100*.Machine$double.eps, ...) {
 	      ## pretest: is it square?
@@ -638,7 +724,7 @@ setMethod("isSymmetric", signature(object = "sparseMatrix"),
 		  ## use gC; "T" (triplet) is *not* unique!
 		  isTRUE(all.equal(.as.dgC.0.factors(  object),
 				   .as.dgC.0.factors(t(object)),
-				   tol = tol, ...))
+				   tolerance = tol, ...))
 	      else if (is(object, "lMatrix"))
 		  ## test for exact equality; FIXME(?): identical() too strict?
 		  identical(as(	 object,  "lgCMatrix"),
@@ -736,9 +822,7 @@ setMethod("cov2cor", signature(V = "sparseMatrix"),
 	      Is <- sqrt(1/diag(V))
 	      if (any(!is.finite(Is))) ## original had 0 or NA
 		  warning("diag(.) had 0 or NA entries; non-finite result is doubtful")
-	      ## TODO: if  <diagonal> %*% <sparse> was implemented more efficiently
-	      ##       we'd rather use that!
-	      Is <- as(Diagonal(x = Is), "sparseMatrix")
+	      Is <- Diagonal(x = Is)
 	      r <- Is %*% V %*% Is
 	      r[cbind(1:p,1:p)] <- 1 # exact in diagonal
 	      as(r, "symmetricMatrix")
@@ -765,106 +849,98 @@ setMethod("all.equal", c(target = "sparseMatrix", current = "sparseMatrix"),
 	  function(target, current, check.attributes = TRUE, ...)
       {
 	  msg <- attr.all_Mat(target, current, check.attributes=check.attributes, ...)
-	  if(is.list(msg)) return(msg[[1]])
-	  ## else
-	  r <- all.equal(as(target, "sparseVector"), as(current, "sparseVector"),
-			 check.attributes=check.attributes, ...)
-	  if(is.null(msg) & (r.ok <- isTRUE(r))) TRUE else c(msg, if(!r.ok) r)
+	  if(is.list(msg)) msg[[1]]
+	  else .a.e.comb(msg,
+			 all.equal(as(target, "sparseVector"), as(current, "sparseVector"),
+				   check.attributes=check.attributes, ...))
       })
 setMethod("all.equal", c(target = "sparseMatrix", current = "ANY"),
 	  function(target, current, check.attributes = TRUE, ...)
       {
 	  msg <- attr.all_Mat(target, current, check.attributes=check.attributes, ...)
-	  if(is.list(msg)) return(msg[[1]])
-	  ## else
-	  r <- all.equal(as(target, "sparseVector"), current,
-			 check.attributes=check.attributes, ...)
-	  if(is.null(msg) & (r.ok <- isTRUE(r))) TRUE else c(msg, if(!r.ok) r)
+	  if(is.list(msg)) msg[[1]]
+	  else .a.e.comb(msg,
+			 all.equal(as(target, "sparseVector"), current,
+				   check.attributes=check.attributes, ...))
       })
 setMethod("all.equal", c(target = "ANY", current = "sparseMatrix"),
 	  function(target, current, check.attributes = TRUE, ...)
       {
 	  msg <- attr.all_Mat(target, current, check.attributes=check.attributes, ...)
-	  if(is.list(msg)) return(msg[[1]])
-	  ## else
-	  r <- all.equal(target, as(current, "sparseVector"),
-			 check.attributes=check.attributes, ...)
-	  if(is.null(msg) & (r.ok <- isTRUE(r))) TRUE else c(msg, if(!r.ok) r)
+	  if(is.list(msg)) msg[[1]]
+	  else .a.e.comb(msg,
+			 all.equal(target, as(current, "sparseVector"),
+				   check.attributes=check.attributes, ...))
       })
 
 
+setMethod("writeMM", "sparseMatrix",
+	  function(obj, file, ...)
+	  writeMM(as(obj, "CsparseMatrix"), as.character(file), ...))
+
 ### --- sparse model matrix,  fac2sparse, etc ----> ./spModels.R
 
+###  xtabs(*, sparse = TRUE) ---> part of standard package 'stats' since R 2.10.0
 
-###--- TODO: Remove, once we require  R >= 2.10.0 ---- :
-
-## xtabs returning a sparse matrix.  This is cut'n'paste
-## of xtabs() in <Rsrc>/src/library/stats/R/xtabs.R ;
-## with the new argument 'sparse'
-xtabs <- function(formula = ~., data = parent.frame(), subset, sparse = FALSE,
-		  na.action, exclude = c(NA, NaN), drop.unused.levels = FALSE)
+##' @title Random Sparse Matrix
+##' @param nrow,
+##' @param ncol number of rows and columns, i.e., the matrix dimension
+##' @param nnz number of non-zero entries
+##' @param rand.x random number generator for 'x' slot
+##' @param ... optionally further arguments passed to sparseMatrix()
+##' @return a sparseMatrix of dimension (nrow, ncol)
+##' @author Martin Maechler
+##' @examples M1 <- rsparsematrix(1000, 20, nnz = 200)
+##'           summary(M1)
+if(FALSE) ## better version below
+rsparsematrix <- function(nrow, ncol, nnz,
+                          rand.x = function(n) signif(rnorm(nnz), 2),
+                          warn.nnz = TRUE, ...)
 {
-    if (missing(formula) && missing(data))
-	stop("must supply either 'formula' or 'data'")
-    if(!missing(formula)){
-	## We need to coerce the formula argument now, but model.frame
-	## will coerce the original version later.
-	formula <- as.formula(formula)
-	if (!inherits(formula, "formula"))
-	    stop("'formula' missing or incorrect")
-    }
-    if (any(attr(terms(formula, data = data), "order") > 1))
-	stop("interactions are not allowed")
-    m <- match.call(expand.dots = FALSE)
-    if (is.matrix(eval(m$data, parent.frame())))
-	m$data <- as.data.frame(data)
-    m$... <- m$exclude <- m$drop.unused.levels <- m$sparse <- NULL
-    m[[1]] <- as.name("model.frame")
-    mf <- eval(m, parent.frame())
-    if(length(formula) == 2) {
-	by <- mf
-	y <- NULL
-    }
-    else {
-	i <- attr(attr(mf, "terms"), "response")
-	by <- mf[-i]
-	y <- mf[[i]]
-    }
-    by <- lapply(by, function(u) {
-	if(!is.factor(u)) u <- factor(u, exclude = exclude)
-	u[ , drop = drop.unused.levels]
-    })
-    if(!sparse) { ## identical to stats::xtabs
-	x <-
-	    if(is.null(y))
-		do.call("table", by)
-	    else if(NCOL(y) == 1)
-		tapply(y, by, sum)
-	    else {
-		z <- lapply(as.data.frame(y), tapply, by, sum)
-		array(unlist(z),
-		      dim = c(dim(z[[1]]), length(z)),
-		      dimnames = c(dimnames(z[[1]]), list(names(z))))
-	    }
-	x[is.na(x)] <- 0
-	class(x) <- c("xtabs", "table")
-	attr(x, "call") <- match.call()
-	x
-
-    } else { ## sparse
-	if (length(by) != 2)
-	    stop("xtabs(*, sparse=TRUE) applies only to two-way tables")
-	rows <- by[[1]]
-	cols <- by[[2]]
-	rl <- levels(rows)
-	cl <- levels(cols)
-	if (is.null(y))
-	    y <- rep.int(1, length(rows))
-	as(new("dgTMatrix",
-	       i = as.integer(rows) - 1L,
-	       j = as.integer(cols) - 1L,
-	       x = as.double(y),
-	       Dim = c(length(rl), length(cl)),
-	       Dimnames = list(rl, cl)), "CsparseMatrix")
+    maxi.sample <- 2^31 # maximum n+1 for which sample(n) returns integer
+    stopifnot((nnz <- as.integer(nnz)) >= 0,
+	      nrow >= 0, ncol >= 0, nnz <= nrow * ncol,
+	      nrow < maxi.sample, ncol < maxi.sample)
+    ## to ensure that nnz is strictly followed, must act on duplicated (i,j):
+    i <- sample.int(nrow, nnz, replace = TRUE)
+    j <- sample.int(ncol, nnz, replace = TRUE)
+    dim <- c(nrow, ncol)
+    it <- 0
+    while((it <- it+1) < 100 &&
+	  anyDuplicated(n.ij <- encodeInd2(i, j, dim, checkBnds=FALSE))) {
+	m <- length(k.dup <- which(duplicated(n.ij)))
+	Matrix.msg(sprintf("%3g duplicated (i,j) pairs", m), .M.level = 2)
+	if(runif(1) <= 1/2)
+	    i[k.dup] <- sample.int(nrow, m, replace = TRUE)
+	else
+	    j[k.dup] <- sample.int(ncol, m, replace = TRUE)
     }
+    if(warn.nnz && it == 100 && anyDuplicated(encodeInd2(i, j, dim, checkBnds=FALSE)))
+	warning("number of non zeros is smaller than 'nnz' because of duplicated (i,j)s")
+    sparseMatrix(i = i, j = j, x = rand.x(nnz), dims = dim, ...)
 }
+
+## No warn.nnz needed, as we sample the encoded (i,j) with*out* replacement:
+rsparsematrix <- function(nrow, ncol, density,
+                          nnz = round(density * maxE), symmetric = FALSE,
+                          rand.x = function(n) signif(rnorm(nnz), 2), ...)
+{
+    maxE <- if(symmetric) nrow*(nrow+1)/2 else nrow*ncol
+    stopifnot((nnz <- as.integer(nnz)) >= 0,
+	      nrow >= 0, ncol >= 0, nnz <= maxE)
+    ## sampling with*out* replacement (replace=FALSE !):
+    ijI <- -1L +
+	if(symmetric) sample(indTri(nrow, diag=TRUE), nnz)
+	else sample.int(maxE, nnz)
+    ## i,j below correspond to  ij <- decodeInd(code, nr) :
+    if(is.null(rand.x))
+	sparseMatrix(i = ijI  %% nrow,
+		     j = ijI %/% nrow,
+		     index1 = FALSE, symmetric = symmetric, dims = c(nrow, ncol), ...)
+    else
+	sparseMatrix(i = ijI  %% nrow,
+		     j = ijI %/% nrow,
+		     index1 = FALSE, symmetric = symmetric,
+		     x = rand.x(nnz), dims = c(nrow, ncol), ...)
+}
+
